/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive;

/**
 * An Agent is a couple (Thread, ReactiveEntity).
 * The life of an Agent presents 3 phases :
 * 1) Initialisation.
 * 2) Communication while a given condition is true.
 * 3) Termination.
 */
public interface Agent {

	/**
	 * Predefined required key arguments of LaunchingContext.
	 * A zero value means no access.
	 */	
	public static final String ARG_L_AGENT_LAUNCHING_KEY =
		"agentLaunchingKey";
	public static final String ARG_P_AGENT_PREVENTING_KEY =
		"agentPreventingKey";
	
	/**
	 * Predefined required arguments of LaunchingContext, other than keys.
	 */
	public static final String ARG_REACTIVE_ENTITY = "reactiveEntity";
	public static final String ARG_AGENT_PROFILE = "agentProfile";
	public static final String ARG_REACTIVE_MACHINE = "reactiveMachine";

	/**
	 * Predefined facultative arguments of LaunchingContext. 
	 */
	public static final String ARG_DESTINATION_COMMUNITY =
		"destinationCommunity";
	public static final String ARG_TRACE_PRINT_STREAM = "tracePrintStream";
	
	/**
	 * Executes the initialisation script.
	 * If the LaunchingContext contains a DESTINATION_COMMUNITY
	 *   then connects to it
	 *   else creates a ReactiveCommunity, connects to it
	 *   and saves it at DESTINATION_COMMUNITY.
	 * @param launchingContext an inherited context for this Agent.
	 *   Predefined required arguments :
	 *     (ARG_L_AGENT_LAUNCHING_KEY, Long)
	 *     (ARG_P_AGENT_PREVENTING_KEY, Long)
	 *     (ARG_REACTIVE_ENTITY, ReactiveEntity)
	 *     (ARG_AGENT_PROFILE, String)
	 *     (ARG_REACTIVE_MACHINE, ReactiveMachine)
	 *   Predefined facultative arguments :
	 *     (ARG_DESTINATION_COMMUNITY, ReactiveCommunity)
	 *     (ARG_TRACE_PRINT_STREAM, PrintStream)
	 */	
	public void init(LaunchingContext launchingContext); 
	
	/**
	 * Returns the persistent Agent identification.
	 * It is the same from launching to launching.
	 * It should be unique inside each frequented community.
	 * Syntax :
	 *   AgentId ::= fullClassName [ '/' parameter ]*
	 *   parameter ::= ( ['-'] [digit]+ ) | ( '"' [char]* '"' )
	 * @return the persistent Agent identification.
	 */
	public String getAgentId();
	
	/**
	 * Returns true if this Agent has been launched, false elsewhere.
	 * @return true if this Agent has been launched, false elsewhere.
	 */
	public boolean isLaunched();
	
	/**
	 * Criterium for continuing to participate to exchanges.
	 * @return true for continuing exchanges, false to quit.
	 */
	public boolean continues();

	static final int INFINITE_IDLE_PERIODE_DURATION = 0;
	
	/**
	 * Returns the duration in millisecond for reacting
	 *   to an idle period. Returns zero for an infinite one.
	 * @return the duration in millisecond for reacting
	 *   to the next idle period. Returns zero for an infinite one.
	 */
	public int idlePeriodDuration_ms();
	
	/**
	 * Oportunity to react to the end of an idle periode.
	 * Pre-condition: the given idle period duration has not been interrupted.
	 */
	public void endOfIdlePeriod();
	
	/**
	 * Processes the given Event and returns true.
	 * Returns false if the Event is unknown.
	 * @param event an Event to process.
	 * @return true if processed, false elsewhere.
	 */
	public boolean processEvent(Event event);

	/**
	 * Processes the given Message and its attachment and returns true.
	 * Returns false if the Message or the attachment content is unknown.
	 * @param message a Message to process.
	 * @param attachment an attachment to the Message.
	 *   It must be not modifiable.
	 * @return true if processed, false elsewhere.
	 */
	public boolean processMessage(Message message, Object attachment);

	/**
	 * Clean-up before disconnect.
	 * Usually owned events are unvalidated here.
	 */
	public void terminate();
}
