/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive;

import org.xmloperator.reactive.exception.BoardAlreadyAttachedException;
import org.xmloperator.reactive.exception.BoardAttachmentException;
import org.xmloperator.reactive.exception.UndefinedEventOrMessageNameException;

/**
 * Community of ReactiveEntity, which are synchronized by instants.
 * Rule : a ReactiveCommunity doesn't publish the reference in memory of
 * 	any ReactiveEntity.
 */
public interface ReactiveCommunity {

	/**
	 * @return the current instant.
	 */
	public long getInstant();
	
	/**
	 * Returns <tt>true</tt> if this ReactiveCommunity has currently no
	 *   connected entity, <tt>false</tt> elsewhere.
	 * Code exemple of the end of a reactive program :
	 * 	<code>while (!community.isIdle()) Thread.yield();</code>
	 * @return <tt>true</tt> if currently idle, i.e. no ReactiveEntity connected.
	 */
	public boolean isIdle();
	
	/**
	 * Returns <tt>true</tt> if this ReacticeCommunity is closed.
	 * Once closed a ReactiveCommunity doesn't accept connections, doesn't
	 *   broadcast any more events and doesn't change any more instant.
	 * @return <tt>true</tt> if this ReacticeCommunity is closed.
	 */
	public boolean isClosed();
	
	/**
	 * Searches an Event or a Message with the given name and returns it.
	 * @param name : an Event name.
	 * @return the Event with the given name. It can be a Message.
	 * @exception UndefinedEventOrMessageNameException is throwed if there no
	 *   Event neither Message with the given name.
	 */
	public Event searchEvent(String name)
		throws UndefinedEventOrMessageNameException;
	
	/**
	 * Returns the Action Message associated to this ReactiveCommunity.
	 * The Action Message belongs to the community and has no name.
	 * @return the Action Message associated to this ReactiveCommunity.
	 */
	public Message getActionMessage();
	
	/**
	 * Returns the Board of this ReactiveCommunity.
	 * In order to assume determinism, writing on the Board is realised
	 *   through Action(s) executed during the inactive part of the instants.
	 * Can be null. Becomes final once not null.
	 * @return the Board of this ReactiveCommunity.
	 */
	public Board getBoard();
	
	/**
	 * Set the Board of this ReactiveCommunity.
	 * This is possible only while the Board is null.
	 * The method getBoard() will return the fixed Board at the next instant.
	 * A Board is read-only during the active part of instants.
	 * A Board cannot be attached to more than one ReactiveCommunity.
	 * @param board a Board.
	 * @param writingKey a key value for restricting access to the methods
	 *   that touches the Elements of the board.
	 * @exception BoardAlreadyAttachedException if a board is already attached
	 *   to this ReactiveCommunity.
	 * @exception BoardAttachmentException if the given Board is attached
	 *   to another ReactiveCommunity.
	 */
	public void setBoard(Board board, long writingKey)
			throws BoardAlreadyAttachedException, BoardAttachmentException;
	
	/**
	 * Returns the theme attached to this ReactiveCommunity.
	 * A theme is used for differentiating communities.
	 * @return the theme attached to this ReactiveCommunity.
	 */
	public String getTheme();
}
