/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive;

import org.xmloperator.reactive.exception.AgentAlreadyLaunchedException;
import org.xmloperator.reactive.exception.BadCharInNameException;
import org.xmloperator.reactive.exception.NameAlreadyUsedException;
import org.xmloperator.reactive.exception.NoMoreValidEventException;
import org.xmloperator.reactive.exception.DisconnectedEntityException;
import org.xmloperator.reactive.exception.NotOwnThreadException;
import org.xmloperator.reactive.exception.EventOwnerException;
import org.xmloperator.reactive.exception.ClosedCommunityException;
import org.xmloperator.reactive.exception.CommunityOwnerException;
import org.xmloperator.reactive.exception.PreventedActionException;

/**
 * A ReactiveEntity is attached to a thread.
 * The thread communicates with a ReactiveCommunity through this interface.
 * The couple (Thread, ReactiveEntity) is called "agent".
 * Rule : an agent must not publish the reference in memory of any
 *   ReactiveEntity.
 */
public interface ReactiveEntity {

	/**
	 * Returns the entityId associated to this ReactiveEntity.
	 * The entityId is internalized and unique within the virtual machine.
	 * @return the entityId associated to this ReactiveEntity.
	 */
	public String getEntityId();
	
	/**
	 * @return the ReactiveCommunity on which it is connected,
	 * 	null if it is not connected.
	 */	
	public ReactiveCommunity getCommunity();
	
	/**
	 * Connects itself to the given community.
	 * If already connected to another community then first disconnect itself
	 *   from it.
	 * Do nothing if already connected to the given community.
	 * The connection occurs during the inactive part of an instant.
	 * This method can only be called by its owner thread.
	 * @param community : the ReactiveCommunity on which to be connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception ClosedCommunityException if the ReactiveCommunity is closed.
	 * @exception PreventedActionException is throwed if this action is not
	 *   allowed for this entity at this time.
	 */
	public void connect(ReactiveCommunity community) throws
			NotOwnThreadException,
			ClosedCommunityException,
			PreventedActionException;
	
	/**
	 * Generates the given event within the current instant.
	 * Pre-condition : The ReactiveEntity must be connected.
	 * This method can only be called by its owner thread.
	 * @param event : an Event to generate. It must be valid.
	 * @exception NoMoreValidEventException if the given Event is no more valid.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception PreventedActionException is throwed if this action is not
	 *   allowed for this entity at this time.
	 */
	public void generateEvent(Event event) throws
			NoMoreValidEventException,
			DisconnectedEntityException,
			NotOwnThreadException,
			PreventedActionException;
	
	/**
	 * Generates a message for broadcasting at the beginning of the next instant.
	 * Pre-condition : The ReactiveEntity must be connected.
	 * This method can only be called by its owner thread.
	 * @param message : a Message to broadcast. It must be valid.
	 * @param attachment : an attachment to the message. It may be null.
	 * @exception NoMoreValidEventException if the given Message is
	 *   no more valid.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception IllegalArgumentException if the attachment doesn't correspond
	 *   to the message.
	 * @exception PreventedActionException is throwed if this action is not
	 *   allowed for this entity at this time.
	 */
	public void generateMessage(Message message, Object attachment) throws
			NoMoreValidEventException,
			DisconnectedEntityException,
			NotOwnThreadException,
			IllegalArgumentException,
			PreventedActionException;

	/**
	 * Generates an Action message for being processed during the next
	 *   instant transition.
	 * Pre-condition : The ReactiveEntity must be connected.
	 * This method can only be called by its owner thread.
	 * @param action : an Action whose status is PROPOSED or EXECUTABLE.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception PreventedActionException is throwed if this action is not
	 *   allowed for this entity at this time.
	 */
	public void generateActionMessage(Action action) throws
			DisconnectedEntityException,
			NotOwnThreadException,
			PreventedActionException;

	/**
	 * To use as argument of the <code>waitEvent()</code> method.
	 */
	static final int AN_EVENT_AT_ANY_TIME = 0;
	
	/**
	 * Waits a given laps of time, which can be interrupted by an Event or
	 *   a connection to the ReactiveCommunity attached to this ReactiveEntity.
	 * @param delay_ms a delay in milliseconds. A zero delay means waiting only
	 *   an Event.
	 * @return true if the delay is used completely, false if interrupted or
	 *   not used at all.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception PreventedActionException is throwed if this action is no more
	 *   allowed for this entity.
	 */
	public boolean waitEvent(int delay_ms) throws
			DisconnectedEntityException,
			NotOwnThreadException,
			PreventedActionException;

	/**
	 * Suspend itself for the current instant.
	 * Pre-condition : The ReactiveEntity must be connected.
	 * It could yet react on events at the current instant.
	 * If some Events occur on the same instant then returns them.
	 * An Event that was generated by itself during the same instant may appear.
	 * 	This occurs when this entity is the first one to generate this Event
	 * 	during this instant.
	 * If some Messages are available then returns them.
	 * If none then stops and returns null.
	 * This method can only be called by its owner thread.
	 * @return some occuring Events, if any.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception PreventedActionException is throwed if this action is no more
	 *   allowed for this entity.
	 */
	public Event[] suspend() throws
			DisconnectedEntityException,
			NotOwnThreadException,
			PreventedActionException;
	
	/**
	 * Stops itself for the current instant.
	 * It will not react any more on the current instant.
	 * Pre-condition : The ReactiveEntity must be connected.
	 * Returns at the beginning of the next instant.
	 * This method can only be called by its owner thread.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception PreventedActionException is throwed if this action is no more
	 *   allowed for this entity.
	 */
	public void stop() throws
			DisconnectedEntityException,
			NotOwnThreadException,
			PreventedActionException;
	
	/**
	 * Disconnect itself from its current ReactiveCommunity.
	 * This method can only be called by its owner thread.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 */
	public void disconnect() throws NotOwnThreadException;
	
	/**
	 * Closes its ReactiveCommunity. Doesn't disconnect.
	 * Must be the creator of the ReactiveCommunity.
	 * Once closed, the community doesn't accept any more connection.
	 * Once idle, the community's automata terminates.
	 * This method can only be called by its owner thread.
	 * Pre-condition : The ReactiveEntity must be connected.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception CommunityOwnerException if this ReactiveEntity is not
	 *   the one that has created the ReactiveCommunity.
	 */
	public void closeCommunity() throws
			DisconnectedEntityException,
			NotOwnThreadException,
			CommunityOwnerException;

	/**
	 * Creates an Event (not a Message) with this as owner and returns it.
	 * Pre-condition : The ReactiveEntity must be connected.
	 * This method can only be called by its owner thread.
	 * @return the created Event.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 */
	public Event createEvent() throws
			DisconnectedEntityException,
			NotOwnThreadException;
	
	/**
	 * Creates a Message with this as owner and returns it.
	 * This cannot be an Action Message.
	 * Pre-condition : The ReactiveEntity must be connected.
	 * This method can only be called by its owner thread.
	 * @param isSenderCertified characteristic of the Message to create.
	 * @return the created Message.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 */
	public Message createMessage(boolean isSenderCertified) throws
			DisconnectedEntityException,
			NotOwnThreadException;
	
	/**
	 * Gives a name to an Event (or a Message).
	 * The name must respect the following rule, which is XML name compatible :
	 * Name ::= ([a-zA-Z] | '_') ([a-zA-Z0-9] | '_' | '-' | '.')*
	 * This method can only be called by its owner thread.
	 * Pre-condition : The event must belong to this ReactiveEntity. 
	 * Pre-condition : The ReactiveEntity must be connected.
	 * Pre-condition : The name is not already an event name (an event name is
	 *   unique).
	 * Post-condition : The event doesn't belong any more to this ReactiveEntity
	 *   but to the ReactiveCommunity.
	 * @param event : an Event to name.
	 * @param name : a name to give to the event. It must be not null.
	 * @exception BadCharInNameException if the given name contains bad chars.
	 * @exception NameAlreadyBoundException if the given name is already used.
	 * @exception NoMoreValidEventException if the given Event is no more valid.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception EventOwnerException if this entity is not the Event owner.
	 * @exception IllegalArgumentException if the given name is null.
	 */
	public void nameEvent(Event event, String name) throws
			BadCharInNameException,
			NameAlreadyUsedException,
			NoMoreValidEventException,
			DisconnectedEntityException,
			NotOwnThreadException,
			EventOwnerException,
			IllegalArgumentException;
	
	/**
	 * Creates an Event with the given name or returns the existing one, if any.
	 * The created Event belongs to the ReactiveCommunity.
	 * The name must respect the following rule, which is XML name compatible :
	 * Name ::= ([a-zA-Z] | '_') ([a-zA-Z0-9] | '_' | '-' | '.')*
	 * Pre-condition : The name is not the one of a Message.
	 * @param name : the name to give to the Event. It must be not null.
	 * @return the created Event.
	 * @exception BadCharInNameException if the given name contains bad chars.
	 * @exception NameAlreadyUsedException if the given name is already used
	 *   by a Message.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception IllegalArgumentException if the given name is null.
	 */
	public Event searchEvent(String name) throws
			BadCharInNameException,
			NameAlreadyUsedException,
			DisconnectedEntityException,
			NotOwnThreadException,
			IllegalArgumentException;
	
	/**
	 * Creates a Message with the given name or returns the existing one, if any.
	 * The created Message belongs to the ReactiveCommunity.
	 * This cannot be an Action Message.
	 * The name must respect the following rule, which is XML name compatible :
	 * Name ::= ([a-zA-Z] | '_') ([a-zA-Z0-9] | '_' | '-' | '.')*
	 * Returns the Message with the given name if it already exists.
	 * Pre-condition : The name is not the one of a base Event or a Message
	 *   with another characteristic than the given one.
	 * @param isSenderCertified characteristic of the Message to search
	 *   or create.
	 * @param name : the name to give to the Message.
	 * @return the created Message.
	 * @exception BadCharInNameException if the given name contains bad chars.
	 * @exception NameAlreadyUsedException if the given name is already used
	 *   by an Event or a Message with another characteristic than the given
	 *     Message.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception IllegalArgumentException if the given name is null.
	 */
	public Message searchMessage(boolean isSenderCertified, String name)
			throws
			BadCharInNameException,
			NameAlreadyUsedException,
			DisconnectedEntityException,
			NotOwnThreadException,
			IllegalArgumentException;
	
	/**
	 * Unvalidates an Event (or a Message).
	 * The event must belong to this ReactiveEntity.
	 * This ReactiveEntity must be connected.
	 * Do nothing if the given Event is not valid. 
	 * This method can only be called by its owner thread.
	 * @param event : an Event to unvalidate.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception EventOwnerException if this entity is not the Event owner.
	 */
	public void unvalidateEvent(Event event) throws
			DisconnectedEntityException,
			NotOwnThreadException,
			EventOwnerException;

	/**
	 * Returns a new ReactiveCommunity.
	 * This ReactiveEntity will the only one to be able to close it.
	 * @param thema a thema to attach to the community. It can be null.
	 * @return a new ReactiveCommunity.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception PreventedActionException is throwed if this action is not
	 *   allowed for this entity at this time.
	 */
	public ReactiveCommunity newCommunity(String thema) throws
			NotOwnThreadException,
			PreventedActionException;

	/**
	 * Launches an Agent into the community.
	 * The launched Agent profile is the same than the launching Agent one.
	 * @param agent : an Agent to launch.
	 * @param context a launching context that will be passed to the Agent.
	 *   It can be null.
	 * @return the entityId of the launched Agent.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception PreventedActionException is throwed if this action is not
	 *   allowed for this entity at this time.
	 * @exception AgentAlreadyLaunchedException if the given Agent is already
	 *   running.
	 */
	public String launchAgent(Agent agent, LaunchingContext context) throws
			DisconnectedEntityException,
			NotOwnThreadException,
			PreventedActionException,
			AgentAlreadyLaunchedException;
}
