/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.element.constraint;

import org.xmloperator.reactive.element.ChildrenConstraint;

/**
 * The ChildrenConstraintFactory build a ChildrenConstraint
 *   from its String value.
 */
public final class ChildrenConstraintFactory {
	
	/**
	 * String value syntax :
	 *   constraint ::= simpleConstraint ( '&' simpleConstraint )*
	 *   simpleConstraint ::= childCountConstraint | nameConstraint
	 *   childCountConstraint ::= 'C' '(' count ',' count? ')'
	 *   nameConstraint ::= 'D' | ( 'F' '(' string ',' string ')' )
	 *   count ::= [0..9]+
	 *   string ::= '"' ([..]-["])+ '"'
	 * @param stringValue
	 * @return
	 */
	public static final ChildrenConstraint childrenConstraint(
			String stringValue) throws IllegalArgumentException {
		return new ChildrenConstraintFactory(stringValue).parseConstraint();
	}
	
	private final char[] chars;
	private final int length;
	private int offset = 0;
	
	private ChildrenConstraintFactory(String stringValue) {
		this.chars = stringValue.toCharArray();
		this.length = this.chars.length;
	}
	
	private ChildrenConstraint parseConstraint() {
		ChildrenConstraint childrenConstraint = this.parseSimpleConstraint();
		while (this.offset < this.length) {
			this.parseChar('&');
			ChildrenConstraint otherConstraint = this.parseSimpleConstraint();
			childrenConstraint =
				new AndChildrenConstraint(childrenConstraint, otherConstraint);
		}
		return childrenConstraint;
	}
	
	private ChildrenConstraint parseSimpleConstraint() {
		if (this.offset == this.length)
			throw new IllegalArgumentException();
		char ch = this.chars[this.offset++];
		switch (ch) {
			case 'C' :
				this.parseChar('(');
				int minCount = this.parseCount();
				if (minCount < 0)
					throw new IllegalArgumentException();
				this.parseChar(',');
				int maxCount = this.parseCount();
				this.parseChar(')');
				return new ChildCountChildrenConstraint(minCount, maxCount);
			case 'D' :
				return DifferentNamesChildrenConstraint.CHILDREN_CONSTRAINT;
			case 'F' :
				this.parseChar('(');
				String namespaceURI = this.parseString();
				this.parseChar(',');
				String localName = this.parseString();
				this.parseChar(')');
				return new FixedNameChildrenConstraint(namespaceURI, localName);
			default :
				throw new IllegalArgumentException();
		}
	}
	
	private void parseChar(char ch) {
		if (this.offset == this.length || this.chars[this.offset++] != ch)
			throw new IllegalArgumentException();
	}
	
	private int parseCount() {
		if (this.offset == this.length)
			return -1;
		char ch = this.chars[this.offset];
		if (ch < '0' || ch > '9')
			return -1;
		int count = (ch - '0');
		this.offset += 1;
		while (true) {
			if (this.offset == this.length)
				break;
			ch = this.chars[this.offset];
			if (ch <'0' || ch > '9')
				break;
			count = 10 * count + (ch - '0');
			this.offset += 1;
		}
		return count;
	}
	
	private String parseString() {
		this.parseChar('"');
		StringBuffer sb = new StringBuffer();
		while (true) {
			if (this.offset == this.length)
				throw new IllegalArgumentException();
			char ch = this.chars[this.offset++];
			if (ch == '"')
				break;
			sb.append(ch);
		}
		return sb.toString();
	}
}
