/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.element.impl;

import java.util.Arrays;
import java.util.Comparator;

import org.xmloperator.reactive.exception.BadCharInNameException;

import org.xmloperator.reactive.element.ChildrenConstraint;
import org.xmloperator.reactive.element.Element;
import org.xmloperator.reactive.element.ParentElement;
import org.xmloperator.reactive.element.RWElement;

/**
 * ParentElement implementation.
 */
public class ParentElementImpl extends ElementImpl
		implements ParentElement {

	protected int childCount = 0;
	protected ElementImpl firstChild = null;
	protected ElementImpl lastChild = null;
	final ChildrenConstraint childrenContraint;

	public ParentElementImpl(String namespaceURI, String localName)
			throws BadCharInNameException {
		super(namespaceURI, localName);
		this.childrenContraint = null;
	}

	public ParentElementImpl(String namespaceURI, String localName,
			Element[] children)
			throws BadCharInNameException {
		super(namespaceURI, localName);
		if (children != null)
			this.addChildren(children);
		this.childrenContraint = null;
	}

	public ParentElementImpl(String namespaceURI, String localName,
			ChildrenConstraint childrenConstraint)
			throws BadCharInNameException {
		super(namespaceURI, localName);
		this.childrenContraint = childrenConstraint;
	}

	public ParentElementImpl(String namespaceURI, String localName,
			Element[] children, ChildrenConstraint childrenConstraint)
			throws BadCharInNameException {
		super(namespaceURI, localName);
		if (children != null)
			this.addChildren(children);
		this.childrenContraint = childrenConstraint;
	}

	final void addChildren(Element[] children) {
		for (int k = 0; k < children.length; k++) {
			ElementImpl child = (ElementImpl)children[k];
			if (child == null)
				throw new IllegalArgumentException(); 
			this.appendChild(child);
		}
	}
	
	RWElement getRWElement() {
		if (this.rwElement == null)
			this.rwElement = new RWParentElementImpl(this);
		return this.rwElement;
	}

	////////////////
	// Write methods
	////////////////

	final void appendChild(ElementImpl element) {
		if (element.parent != null || element == this.root())
			throw new IllegalArgumentException();
		if (this.childrenContraint != null)
			this.childrenContraint.verifyConstraint(this, element, true);
		element.parent = this;
		this.childCount += 1;
		if (this.lastChild != null) {
			this.lastChild.nextSibling = element;
			element.previousSibling = this.lastChild;
		}
		else
			this.firstChild = element;
		this.lastChild = element;
	}
	
	final void removeChildren() {
		while (this.firstChild != null)
			this.firstChild.remove();
	}
	
	final void sortChildren(Comparator elementComparator) {
		ElementImpl[] children = new ElementImpl[this.childCount];
		{
			ElementImpl child = this.firstChild;
			for (int k = 0; k < this.childCount; k++) {
				children[k] = child;
				child = child.nextSibling;
			}
		}
		Arrays.sort(children, elementComparator);
		this.removeChildren();
		for (int k = 0; k < this.childCount; k++)
			this.appendChild(children[k]);
	}
	
	///////////////////////////////
	// ParentElement implementation
	///////////////////////////////
	
	public final int getChildCount() {
		return this.childCount;
	}

	public final Element getFirstChild() {
		return this.firstChild;
	}

	public final Element getLastChild() {
		return this.lastChild;
	}

	public final Element getChild(String namespaceURI, String localName) {
		ElementImpl foundElement = null;
		ElementImpl child = this.firstChild;
		while (child != null) {
			if (child.namespaceURI == namespaceURI
					&& child.localName.equals(localName)) {
				if (foundElement != null)
					return null;
				foundElement = child;
			}
			child = child.nextSibling;
		}
		return foundElement;
	}
	
	public final ChildrenConstraint getChildrenContraint() {
		return this.childrenContraint;
	}
	
	/////////////////////////
	// Element implementation
	/////////////////////////
	
	public boolean isParentElement() {
		return true;
	}

	public boolean equals(Element element) {
		if (element == this)
			return true;
		if (!(element instanceof ParentElementImpl)
				|| !super.equals(element))
			return false;
		ParentElementImpl otherElement = (ParentElementImpl)element;
		if (this.childCount != otherElement.childCount)
			return false;
		ElementImpl thisChild = this.firstChild;
		ElementImpl otherChild = otherElement.firstChild;
		while (thisChild != null) {
			if (!thisChild.equals(otherChild))
				return false;
			thisChild = thisChild.nextSibling;
			otherChild = otherChild.nextSibling;
		}
		return true;
	}

	////////////////////////
	// Object implementation
	////////////////////////
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append(super.toString());
		sb.append('(');
		ElementImpl child = this.firstChild;
		while (child != null) {
			if (child != this.firstChild)
				sb.append(',');
			sb.append(child);
			child = child.nextSibling;
		}
		sb.append(')');
		return sb.toString();
	}
}
