/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.facet;

import java.util.Collections;
import java.util.Set;
import java.util.HashSet;

import org.xmloperator.reactive.exception.ReactiveException;
import org.xmloperator.reactive.exception.BadCharInNameException;

import org.xmloperator.reactive.Terms;
import org.xmloperator.reactive.TermsAttachment;
import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.Agent;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.ReactiveMachine;

import org.xmloperator.reactive.util.MessageAttachment;

/**
 * This facet makes some rules applied in a community, by preventing the agents
 *   that violate them.
 * Rule 1 : A "reject" message must be generated at the instant of reception
 *   of the original message.
 * Rule 2 : A "reject" message must include the original message, without change.
 * Rule 3 : An Agent that emits a "reject" message must have
 *   the PROFILE_AGENT_PREVENTING profile.
 * Rule 4 : An Agent must not be the cause of a valid "reject" message.
 * Application of rule 1, 2, 3 : the rejecting Agent is prevented.
 * Application of rule 4 : the rejected Agent is prevented.
 * Of course, the Agent that contains this facet must have
 *   the PROFILE_AGENT_PREVENTING profile.
 */
public final class AgentPreventFacet extends FacetImpl {

	private static final String REJECT = "reject".intern();
	private static final String PREVENT = "prevent".intern();
	
	static {
		BadCharInNameException.verifyName(REJECT);
		BadCharInNameException.verifyName(PREVENT);
	}
	
	private ReactiveMachine reactiveMachine = null;
	private long agentPreventingKey = 0;
	private Set instantMinus1MessageAttachmentSet = Collections.EMPTY_SET;
	private Set instantMessageAttachmentSet = new HashSet();
	
	public AgentPreventFacet() {
	}
	
	private boolean hasAgentPreventingProfile(String agentId) {
		String agentProfile = this.reactiveMachine.getAgentProfile(agentId);
		return agentProfile != null
			&& agentProfile.indexOf(ReactiveMachine.PROFILE_AGENT_PREVENTING)
				>= 0;
	}
	
	private void preventAgent(String agentId, boolean isRejectAuthor) {
		if (this.trace != null)
			this.trace.println(this.reac.getEntityId() + " prevents " + agentId
					+ (isRejectAuthor
							? " that has generated an unvalid reject"
							: " whose a message has been correctly rejected"));
		this.reactiveMachine.preventAgent(
				this.agentPreventingKey, agentId, false);
	}
	
	////////////////////////////
	// Facet implementation
	////////////////////////////

	public Object getService() {
		return PREVENT;
	}
	
	public void init(LaunchingContext launchingContext) {
		super.init(launchingContext);
		{
			String agentProfile = (String)launchingContext.getArgumentValue(
					Agent.ARG_AGENT_PROFILE);
			if (agentProfile.indexOf(ReactiveMachine.PROFILE_AGENT_PREVENTING)
					< 0)
				throw new ReactiveException(
					"This agent has not the profile for preventing agents");
		}
		this.reactiveMachine =
			(ReactiveMachine)launchingContext.getArgumentValue(
					Agent.ARG_REACTIVE_MACHINE);
		this.agentPreventingKey =
			((Long)this.launchingContext.getArgumentValue(
				Agent.ARG_P_AGENT_PREVENTING_KEY)).longValue();
	}

	public void startOfInstant() {
		this.instantMinus1MessageAttachmentSet =
			this.instantMessageAttachmentSet;
		this.instantMessageAttachmentSet = new HashSet();
	}
	
	public void processServiceExcludedMessage(Message message,
			TermsAttachment attachment) {
		String messageName = message.getName();
		if (messageName == REJECT) {
			String rejecterAgentId = attachment.getSenderEntityId();
			Terms terms = attachment.getTerms();
			boolean isRejectOk = (terms.count() == 2)
					&& (terms.get(0) instanceof Message)
					&& (terms.get(1) instanceof TermsAttachment)
					&& this.hasAgentPreventingProfile(rejecterAgentId); 
			if (isRejectOk) {
				Message rejectedMessage = (Message)terms.get(0);
				TermsAttachment rejectedAttachment =
					(TermsAttachment)terms.get(1);
				isRejectOk &= this.instantMinus1MessageAttachmentSet.contains(
							new MessageAttachment(rejectedMessage,
									rejectedAttachment));
				if (isRejectOk) {
					String rejectedAgentId =
						rejectedAttachment.getSenderEntityId();
					this.preventAgent(rejectedAgentId, false);
				}
			}
			if (!isRejectOk)
				this.preventAgent(rejecterAgentId, true);
		}
		else
			this.instantMessageAttachmentSet.add(
				new MessageAttachment(message, attachment));
	}
}
