/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.facet;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.xmloperator.reactive.exception.BadCharInNameException;

import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.Terms;
import org.xmloperator.reactive.TermsAttachment;

import org.xmloperator.reactive.util.TermsAttachmentFactory;

/**
 * ---------------------------------------------------------------------------
 * This facet is no more used since thema is becomed an attribute
 *   of ReactiveCommunity.
 * ---------------------------------------------------------------------------
 * Using this Facet, the creator of a community can publish/unpublish it.
 * The procols of the "community" service are following :
 *   accept(publish (community [community] [thema]))
 *    -> rejected if the community is attached to another sender agentId.
 *   accept(unpublish (community [community] [thema]))
 *    -> rejected if the community is attached to another sender agentId.
 *   accept(unpublish (community [community]))
 *    -> rejected if the community is attached to another sender agentId.
 *   accept(search (community [thema]))
 *     -> emit((publish (community [community] [thema]))*)
 */
public final class CommunityCatalogFacet extends FacetImpl {

	private static final String PUBLISH = "publish".intern();
	private static final String UNPUBLISH = "unpublish".intern();
	private static final String SEARCH = "search".intern();
	private static final String REJECT = "reject".intern();
	
	private static final String COMMUNITY = "community".intern();

	static {
		BadCharInNameException.verifyName(PUBLISH);
		BadCharInNameException.verifyName(UNPUBLISH);
		BadCharInNameException.verifyName(SEARCH);
		BadCharInNameException.verifyName(REJECT);
		
		BadCharInNameException.verifyName(COMMUNITY);
	}
	
	private final Map fromCommunityToAgentId = new HashMap();	
	private final Map fromThemaToCommunityList = new HashMap();
	private final Map fromCommunityToThemaList = new HashMap();

	public CommunityCatalogFacet() {
	}

	/**
	 * Associates the given ReactiveCommunity and the given thema.
	 * This is allowed only by the community administrator, whose
	 *   agentId is given.
	 * @param agentId the agentId of the administrator of the community.
	 * @param community a ReactiveCommunity.
	 * @param thema a thema.
	 * @return if the association is allowed, false elsewhere.
	 */
	public boolean associate(Object agentId,
			ReactiveCommunity community, String thema) {
		Object adminAgentId = this.fromCommunityToAgentId.get(community);
		if (adminAgentId == null)
			this.fromCommunityToAgentId.put(community, agentId);
		else if (agentId != adminAgentId)
			return false;
		if (!this.areAssociated(community, thema)) {
			this.associate_direct(thema, community);
			this.associate_inverse(thema, community);
			if (this.trace != null)
				this.trace.println(agentId + " associates " + community
						+ " and " + thema);
		}
		return true;
	}
	
	/**
	 * Disassociates the given ReactiveCommunity and the given thema.
	 * This is allowed only by the community administrator, whose
	 *   agentId is given.
	 * @param agentId the agentId of the administrator of the community.
	 * @param community a ReactiveCommunity.
	 * @param thema a thema.
	 * @return if the disassociation is allowed, false elsewhere.
	 */
	public boolean disassociate(Object agentId,
			ReactiveCommunity community, String thema) {
		if (this.fromCommunityToAgentId.get(community) != agentId)
			return false;
		this.disassociate_direct(thema, community);
		this.disassociate_inverse(thema, community);
		if (!this.fromCommunityToThemaList.containsKey(community))
			this.fromCommunityToAgentId.remove(community);
		if (this.trace != null)
			this.trace.println(agentId + " disassociates " + community
					+ " and " + thema);
		return true;
	}
	
	/**
	 * Disassociates the given ReactiveCommunity from any thema.
	 * This is allowed only by the community administrator, whose
	 *   agentId is given.
	 * @param agentId the agentId of the administrator of the community.
	 * @param community a ReactiveCommunity.
	 * @return if the disassociation is allowed, false elsewhere.
	 */
	public boolean disassociate(Object agentId, ReactiveCommunity community) {
		if (this.fromCommunityToAgentId.get(community) != agentId)
			return false;
		List themaList = (List)this.fromCommunityToThemaList.get(community);
		if (themaList != null) {
			Iterator iterator = themaList.iterator();
			while (iterator.hasNext()) {
				String thema = (String)iterator.next();
				this.disassociate_direct(thema, community);
			}
			}
		this.fromCommunityToThemaList.remove(community);
		this.fromCommunityToAgentId.remove(community);
		if (this.trace != null)
			this.trace.println(agentId + " disassociates " + community);
		return true;
	}
	
	private boolean areAssociated(ReactiveCommunity community, String thema) {
		List communityList = (List)this.fromThemaToCommunityList.get(thema);	
		if (communityList == null)
			return false;
		if (!communityList.contains(community))
			return false;
		List themaList = (List)this.fromCommunityToThemaList.get(community);
		if (themaList == null)
			return false;
		if (!themaList.contains(thema))
			return false;
		return true;
	}
	
	private void associate_direct(String thema, ReactiveCommunity community) {
		List communityList = (List)this.fromThemaToCommunityList.get(thema);	
		if (communityList == null) {
			communityList = new ArrayList(4);
			this.fromThemaToCommunityList.put(thema, communityList);	
		}
		if (!communityList.contains(community))
			communityList.add(community);
	}
	
	private void associate_inverse(String thema, ReactiveCommunity community) {
		List themaList = (List)this.fromCommunityToThemaList.get(community);
		if (themaList == null) {
			themaList = new ArrayList(4);
			this.fromCommunityToThemaList.put(community, themaList);	
		}
		if (!themaList.contains(thema))
			themaList.add(thema);
	}
	
	private void disassociate_direct(String thema, ReactiveCommunity community) {
		List communityList = (List)this.fromThemaToCommunityList.get(thema);
		if (communityList != null) {
			communityList.remove(community);
			if (communityList.isEmpty())
				this.fromThemaToCommunityList.remove(thema);
		}
	}
	
	private void disassociate_inverse(String thema, ReactiveCommunity community) {
		List themaList = (List)this.fromCommunityToThemaList.get(community);
		if (themaList != null) {
			themaList.remove(thema);
			if (themaList.isEmpty())
				this.fromCommunityToThemaList.remove(community);	
		}
	}
	
	private boolean process_publish_message(TermsAttachment attachment) {
		Terms terms = attachment.getTerms();
		if (terms.count() != 3)
			return false;
		String senderAgentId = attachment.getSenderEntityId();
		Object service = terms.get(0);
		if (service == COMMUNITY) {
			Object term1 = terms.get(1);
			Object term2 = terms.get(2);
			if (!(term1 instanceof ReactiveCommunity)
					|| !(term2 instanceof String))
				return false;
			{
				ReactiveCommunity community = (ReactiveCommunity)term1;
				String thema = (String)term2;
				try {
					BadCharInNameException.verifyName(thema);
					if (this.associate(senderAgentId, community, thema))
						return true;
				}
				catch (BadCharInNameException exception) {}
			}
			Message m_publish = this.reac.searchMessage(true, PUBLISH);
			Message m_reject = this.reac.searchMessage(true, REJECT);
			this.reac.generateMessage(m_reject,
					TermsAttachmentFactory.newTermsAttachment(
							m_publish, attachment));
			return true;
		}
		return false;
	}

	private boolean process_unpublish_message(TermsAttachment attachment) {
		Terms terms = attachment.getTerms();
		if (terms.count() < 2 || terms.count() > 3)
			return false;
		Object senderAgentId = attachment.getSenderEntityId();
		Object service = terms.get(0);
		if (service == COMMUNITY) {
			Object term1 = terms.get(1);
			if (!(term1 instanceof ReactiveCommunity))
				return false;
			{
				ReactiveCommunity community = (ReactiveCommunity)term1;
				if (terms.count() == 2) {
					if (this.disassociate(senderAgentId, community))
						return true;
				}
				else {
					Object term2 = terms.get(2);
					if (!(term2 instanceof String))
						return false;
					String thema = (String)term2;
					if (this.disassociate(senderAgentId, community, thema))
						return true;
				}
			}
			Message m_unpublish = this.reac.searchMessage(true, UNPUBLISH);
			Message m_reject = this.reac.searchMessage(true, REJECT);
			this.reac.generateMessage(m_reject,
					TermsAttachmentFactory.newTermsAttachment(
							m_unpublish, terms));
			return true;
		}
		return false;
	}

	private boolean process_search_message(TermsAttachment attachment) {
		Terms terms = attachment.getTerms();
		if (terms.count() != 2)
			return false;
		Object service = terms.get(0);
		if (service == COMMUNITY) {
			Object term1 = terms.get(1);
			if (!(term1 instanceof String))
				return false;
			String thema = (String)term1;
			try {
				BadCharInNameException.verifyName(thema);
				List communityList =
					(List)this.fromThemaToCommunityList.get(thema);
				if (communityList != null) {
					Iterator iterator = communityList.iterator();
					while (iterator.hasNext()) {
						ReactiveCommunity community =
							(ReactiveCommunity)iterator.next();
						Message m_publish =
							this.reac.searchMessage(true, PUBLISH);
						this.reac.generateMessage(m_publish,
								TermsAttachmentFactory.newTermsAttachment(
										COMMUNITY, community, thema));
					}
				}
			}
			catch (BadCharInNameException exception) {
				Message m_search =
					this.reac.searchMessage(true, SEARCH);
				Message m_reject =
					this.reac.searchMessage(true, REJECT);
				this.reac.generateMessage(m_reject,
						TermsAttachmentFactory.newTermsAttachment(
								m_search, terms));
			}
			return true;
		}
		return false;
	}

	////////////////////////////
	// Facet implementation
	////////////////////////////

	public Object getService() {
		return COMMUNITY;
	}
	
	public boolean processServiceMessage(
			String messageName, TermsAttachment attachment) {
		if (messageName == PUBLISH) {
			this.process_publish_message(attachment);
			return true;
		}
		if (messageName == UNPUBLISH) {
			this.process_unpublish_message(attachment);
			return true;
		}
		if (messageName == SEARCH) {
			this.process_search_message(attachment);
			return true;
		}
		return false;
	}
}
