/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.impl;

import java.io.PrintStream;
import java.util.Map;
import java.util.HashMap;
import java.util.Random;

import org.xmloperator.reactive.exception.ActiveReactiveMachineException;
import org.xmloperator.reactive.exception.AgentAlreadyLaunchedException;
import org.xmloperator.reactive.exception.DoneMachineException;
import org.xmloperator.reactive.exception.PrivilegiedAccessException;

import org.xmloperator.reactive.Agent;
import org.xmloperator.reactive.ReactiveMachine;
import org.xmloperator.reactive.MachineKeys;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.ReactiveCommunity;
//import org.xmloperator.reactive.ReactiveEntity;
import org.xmloperator.reactive.CrashingReport;

/**
 * ReactiveMachine implementation.
 */
final class ReactiveMachineImpl
		implements ReactiveMachine, ConnectionAcknowledgment {

	private static ReactiveMachine currentlyActiveMachine = null;
	private static PrintStream trace = null;
	private static boolean isFullTrace = false;
	private static boolean isAgentTrace = false;
	
	static void setTrace(PrintStream trace, boolean isFullTrace,
			boolean isAgentTrace) {
		ReactiveMachineImpl.trace = trace;
		ReactiveMachineImpl.isFullTrace = isFullTrace;
		ReactiveMachineImpl.isAgentTrace = isAgentTrace;
	}

	static ReactiveMachine getCurrentlyActiveMachine() {
		return currentlyActiveMachine;
	}
	
	private final Object synchroActiveMachine = new Object();
	private final Object synchroMachineTermination = new Object();
	private final Random random = new Random(); 
	private final Map fromEntityIdToEntity = new HashMap();
	private final Map fromEntityIdToAgentProfile = new HashMap();
	private final Map fromThemaToCommunity = new HashMap();
	private final long agentLaunchingKey = this.newKey();
	private final long agentPreventingKey = this.newKey();
	private final long machineDoneKey = this.newKey();
	private final int breathe_duration_ms;
	private final CrashingReport crashingReport;
	private int aliveAgentCount = 0;
	private int aliveCommunityCount = 0;
	private boolean isLaunchingAgent = false;
	private boolean isAgentLaunched = false;
	private boolean isDone = false;

	public ReactiveMachineImpl(int breathe_duration_ms,
			CrashingReport report, MachineKeys machineKeys) {
		boolean isActiveMachine;
		synchronized (this.synchroActiveMachine) {
			isActiveMachine = currentlyActiveMachine != null;
			if (!isActiveMachine)
				currentlyActiveMachine = this;
		}
		if (isActiveMachine)
			throw new ActiveReactiveMachineException();
		this.breathe_duration_ms = breathe_duration_ms;
		this.crashingReport = report;
		machineKeys.agentLaunchingKey = this.agentLaunchingKey;
		machineKeys.agentPreventingKey = this.agentPreventingKey;
		machineKeys.machineDoneKey = this.machineDoneKey;
	}
	
	/**
	 * Notification of the starting of the given Agent.
	 * @param entityId the entityId of an Agent that is launched.
	 */
	void notifyAgentStarting(String entityId) {
		if (trace != null && isFullTrace)
			trace.println(entityId + " starts");
		this.aliveAgentCount += 1;
		this.isAgentLaunched = true;
	}
	
	/**
	 * Notification of the termination of the given Agent.
	 * @param entityId the entityId of an Agent that terminates.
	 */
	void notifyAgentTermination(String entityId) {
		if (trace != null && isFullTrace)
			trace.println(entityId + " terminates");
		synchronized (this) {
			this.fromEntityIdToAgentProfile.remove(entityId);
			this.fromEntityIdToEntity.remove(entityId);
		}
		synchronized (this.synchroMachineTermination) {
			this.aliveAgentCount -= 1;
			if (this.aliveAgentCount == 0 && this.aliveCommunityCount == 0)
				this.synchroMachineTermination.notifyAll();
		}
	}
	
	/**
	 * Notification of the starting of the given ReactiveCommunity.
	 * @param community the ReactiveCommunity that is created.
	 */
	void notifyCommunityStarting(ReactiveCommunity community) {
		if (trace != null && isFullTrace)
			trace.println(community + " starts");
		this.aliveCommunityCount += 1;
		String thema = community.getTheme();
		if (thema != null)
			synchronized (this) {
				if (!this.fromThemaToCommunity.containsKey(thema))
					this.fromThemaToCommunity.put(thema, community);
			}
	}
	
	/**
	 * Notification of the termination of the given ReactiveCommunity.
	 * @param community the ReactiveCommunity that terminates.
	 */
	synchronized void notifyCommunityTermination(
			ReactiveCommunity community) {
		if (trace != null && isFullTrace)
			trace.println(community + " terminates");
		String thema = community.getTheme();
		if (thema != null)
			synchronized (this) {
				if (this.fromThemaToCommunity.get(thema) == community)
					this.fromThemaToCommunity.remove(thema);
			}
		synchronized (this.synchroMachineTermination) {
			this.aliveCommunityCount -= 1;
			if (this.aliveCommunityCount == 0 && this.aliveAgentCount == 0)
				this.synchroMachineTermination.notifyAll();
		}
	}
	
	/////////////////////////////////////////
	// ConnectionAcknowledgment implemetation
	/////////////////////////////////////////
	
	public void acknowledgeConnection(String entityId) {
		this.notifyAgentStarting(entityId);
	}
	
	/////////////////////////////////
	// ReactiveMachine implementation
	/////////////////////////////////
	
	public String launchAgent(
			long agentLaunchingKey,
			Agent agent,
			LaunchingContext launchingContext,
			String agentProfile) {
		if (agent == null || agentProfile == null)
			throw new IllegalArgumentException();
		if (agentLaunchingKey != this.agentLaunchingKey)
			throw new PrivilegiedAccessException();
		if (this.isDone)
			throw new DoneMachineException();
		boolean isLaunchingAgent;
		do {
			synchronized (this) {
				isLaunchingAgent = this.isLaunchingAgent;
				this.isLaunchingAgent = true;
			}
			if (isLaunchingAgent)
				this.breathe("AgentLaunchingBeDone");
		}
		while (isLaunchingAgent);
		if (agent.isLaunched()) {
			this.isLaunchingAgent = false;
			throw new AgentAlreadyLaunchedException();
		}
		if (launchingContext == null)
			launchingContext = new LaunchingContextImpl();
		if (trace != null && isAgentTrace)
			launchingContext.addArgument(Agent.ARG_TRACE_PRINT_STREAM, trace);
		launchingContext.addArgument(Agent.ARG_AGENT_PROFILE, agentProfile);
		launchingContext.addArgument(Agent.ARG_REACTIVE_MACHINE, this);
		{
			boolean isAgentLaunchingKey = agentProfile.indexOf(
					ReactiveMachine.PROFILE_AGENT_LAUNCHING) >= 0;
			launchingContext.addArgument(Agent.ARG_L_AGENT_LAUNCHING_KEY,
					new Long(isAgentLaunchingKey ? agentLaunchingKey : 0));
		}
		{
			boolean isAgentPreventingKey = agentProfile.indexOf(
					ReactiveMachine.PROFILE_AGENT_PREVENTING) >= 0;
			launchingContext.addArgument(Agent.ARG_P_AGENT_PREVENTING_KEY,
					new Long(isAgentPreventingKey ? this.agentPreventingKey
							: 0));
		}
		ReactiveEntityImpl reac =
			new ReactiveEntityImpl(this, agentProfile, agentLaunchingKey);
		launchingContext.addArgument(Agent.ARG_REACTIVE_ENTITY, reac);
		String entityId = reac.getEntityId();
		this.fromEntityIdToEntity.put(entityId, reac);
		this.fromEntityIdToAgentProfile.put(entityId, agentProfile);
		AgentThread agentThread = new AgentThread(
				this,
				agent,
				agentProfile,
				launchingContext,
				this.crashingReport);
		agentThread.start();
		Thread.yield();
		while (!this.isAgentLaunched)
			this.breathe("AgentBeLaunched");
		this.isAgentLaunched = false;
		this.isLaunchingAgent = false;
		return entityId;
	}

	public long newKey() {
		while (true) {
			long key = this.random.nextLong() + new Random().nextLong();
			if (key != 0)
				return key;
		}
	}

	public String getAgentProfile(String entityId) {
		return (String)this.fromEntityIdToAgentProfile.get(entityId);
	}
	
	public void preventAgent(
			long agentPreventingKey,
			String entityId,
			boolean concernsAnyAccess) {
		if (entityId == null)
			throw new IllegalArgumentException();
		if (agentPreventingKey != this.agentPreventingKey)
			throw new PrivilegiedAccessException();
		if (this.isDone)
			throw new DoneMachineException();
		ReactiveEntityImpl reactiveEntity =
				(ReactiveEntityImpl)this.fromEntityIdToEntity.get(entityId);
		if (reactiveEntity == null)
			return;
		reactiveEntity.prevent(concernsAnyAccess);
		if (trace != null)
			trace.println(entityId + " is prevented"
				+ (concernsAnyAccess ? " for any access" : ""));
	}
	
	public ReactiveCommunity getCommunity(String thema) {
		if (thema == null)
			throw new IllegalArgumentException();
		return (ReactiveCommunity)this.fromThemaToCommunity.get(thema);
	}
	
	public void done(long agentLauncherDoneKey) {
		if (this.isDone)
			return;
		if (agentLauncherDoneKey != this.machineDoneKey)
			throw new PrivilegiedAccessException();
		/*
		 * Waiting that all agents terminate and all communities has been closed.
		 */
		if (trace != null)
			trace.println("ReactiveMachine waits for "
					+ this.aliveCommunityCount + " community(s) and "
					+ this.aliveAgentCount + " agent(s) be done");
		try {
			synchronized (this.synchroMachineTermination) {
				if (this.aliveAgentCount > 0 || this.aliveCommunityCount > 0)
					this.synchroMachineTermination.wait();
			}
		}
		catch (InterruptedException exception) {}
		/*
		 * Terminates the ReactiveMachine.
		 */
		this.isDone = true;
		currentlyActiveMachine = null;
		if (trace != null)
			trace.println("ReactiveMachine is done");
	}
	
	/////////////////////////
	// Machine implementation
	/////////////////////////
	
	public void breathe(String motivation) {
		try {
			Thread.sleep(this.breathe_duration_ms);
		}
		catch (InterruptedException exception) {}
		if (trace != null && isFullTrace)
			trace.println(Thread.currentThread().getName()
					+ " breathe for " + motivation);
	}
}
